﻿
using System;
using Ozeki.Common;
using Ozeki.Media;
using Ozeki.Media.MediaHandlers;
using Ozeki.MediaGateway;
using Ozeki.MediaGateway.Data;
using Ozeki.MediaGateway.Service;
using Ozeki.Network.Nat;
using Ozeki.VoIP;
using Ozeki.VoIP.Media;
using Ozeki.VoIP.SDK;
using System.Threading;

namespace SIPGateway
{
    class SIPClient
    {
        IClient client;

        ISoftPhone softPhone;
        IPhoneLine phoneLine;
        IPhoneCall call;

        IStreamService streamService;
        string publishStreamName;
        string playStreamName;

        MediaConnector mediaConnector;
        PhoneCallAudioSender phoneCallAudioSender;
        PhoneCallAudioReceiver phoneCallAudioReceiver;

        MediaGatewayAudioReceiver mediaGatewayAudioReceiver;
        MediaGatewayAudioSender mediaGatewayAudioSender;

        /// <summary>
        /// Creates a new instance of SIPClient.
        /// </summary>
        /// <param name="client">The connected client.</param>
        /// <param name="streamService"> Stream service what transmits media data between two client.</param>
        /// <param name="softPhone"> Represents a SIP softphone</param>
        /// <param name="account">SIP account for PBX registration.</param>
        public SIPClient(IClient client, IStreamService streamService, ISoftPhone softPhone, SIPAccount account)
        {
            mediaConnector = new MediaConnector();
            phoneCallAudioSender = new PhoneCallAudioSender();
            phoneCallAudioReceiver = new PhoneCallAudioReceiver();

            this.streamService = streamService;
            this.softPhone = softPhone;
            this.client = client;

            Account = account;
        }

        /// <summary>
        /// SIP account for able to register a SIP PBX.
        /// </summary>
        public SIPAccount Account { get; protected set; }

        /// <summary>
        /// Indicates phoneline information changes.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void PhoneLine_PhoneLineInformationCanged(object sender, VoIPEventArgs<PhoneLineState> e)
        {
            if (e.Item == PhoneLineState.RegistrationSucceeded || e.Item == PhoneLineState.NoRegNeeded)
                client.InvokeMethod("OnRegisteredUserNameReceived", Account.DisplayName);


            OnPhoneLineStateChanged(e.Item);
        }

        /// <summary>
        /// Indicates received incoming call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void SoftPhone_IncomingCall(object sender, VoIPEventArgs<IPhoneCall> e)
        {

            if (e.Item.PhoneLine == phoneLine) 
            {
                call = e.Item;
                call.CallErrorOccured += Call_ErrorOccured;
                call.CallStateChanged += Call_StateChanged;
                call.DtmfReceived += call_DtmfReceived;
                OnIncomingCall(call.DialInfo.DisplayName);
            }
        }


        /// <summary>
        /// Call the given dial number.
        /// </summary>
        /// <param name="dialNumber">dial number</param>
        public virtual void Call(string dialNumber)
        {
            if (phoneLine == null)
            {
                System.Diagnostics.Debug.Fail("PhoneLine null");
                return;
            }

            if ((phoneLine.RegisteredInfo == PhoneLineState.RegistrationSucceeded || phoneLine.RegisteredInfo == PhoneLineState.NoRegNeeded) && call == null)
            {
                call = softPhone.CreateCallObject(phoneLine, dialNumber);

                call.CallErrorOccured += Call_ErrorOccured;
                call.CallStateChanged += Call_StateChanged;
                call.DtmfReceived += call_DtmfReceived;
                call.Start();
            }
        }

        /// <summary>
        /// Indicates call state changes.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected virtual void Call_StateChanged(object sender, VoIPEventArgs<CallState> e)
        {
            switch (e.Item)
            {
                case CallState.InCall:
                  phoneCallAudioSender.AttachToCall(call);
                    phoneCallAudioReceiver.AttachToCall(call);

                    publishStreamName = Guid.NewGuid().ToString(); // Ez az amit a kliens publikál
                    playStreamName = Guid.NewGuid().ToString(); // Ez az amit a szerver publikál

                    var mediaStream = streamService.CreateStream(playStreamName);
                    mediaGatewayAudioReceiver = new MediaGatewayAudioReceiver(mediaStream);

                    mediaConnector.Connect(phoneCallAudioReceiver, mediaGatewayAudioReceiver);

                    OnInCall(publishStreamName, playStreamName);
                    break;

                case CallState.Completed:
                    phoneCallAudioSender.Detach();
                    phoneCallAudioReceiver.Detach();
                    
                    streamService.RemoveStream(mediaGatewayAudioReceiver.MediaStream);
                    mediaConnector.Disconnect(phoneCallAudioReceiver, mediaGatewayAudioReceiver);
                    mediaGatewayAudioReceiver.Dispose();

                    mediaGatewayAudioReceiver = null;
                    call = null;
                    break;
            }
            OnCallStateChanged(e.Item);
        }

        /// <summary>
        /// Indicates call errors. For example : The called number can not be found, equal with Callstate.NotFound
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected virtual void Call_ErrorOccured(object sender, VoIPEventArgs<CallError> e)
        {
            OnCallErrorOccurred(e.Item);
        }

        /// <summary>
        /// Indicates receiving DTMF signals.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"> The tpy</param>
        protected void call_DtmfReceived(object sender, VoIPEventArgs<OzTuple<VoIPMediaType, DtmfSignal>> e)
        {
            OnCallDTMFReceived(e.Item.Item2);
        }

        /// <summary>
        /// Hangup the call.
        /// </summary>
        public virtual void HangUp()
        {
            if (call != null)
            {
                call.HangUp();
                call = null;
            }
        }

        /// <summary>
        /// Accept the incoming call request.
        /// </summary>
        public virtual void Accept()
        {
            if (call != null)
            {
                call.Accept();
            }
        }
        /// <summary>
        /// Reject the incoming call request.
        /// </summary>
        public void Reject()
        {
            if (call != null)
            {
                call.Reject();
                call = null;
            }
        }

        /// <summary>
        /// Stats to send a DTMF sign.
        /// </summary>
        /// <param name="dtmf">DTMF sign</param>
        public void StartDtmf(DtmfNamedEvents dtmf)
        {
            if (call != null)
                call.StartDTMFSignal(VoIPMediaType.Audio, dtmf);
        }

        /// <summary>
        /// Stops to send a DTMF sign.
        /// </summary>
        /// <param name="dtmf"></param>
        public void StopDtmf(DtmfNamedEvents dtmf)
        {
            if (call != null)
                call.StopDTMFSignal(VoIPMediaType.Audio, dtmf);
        }

		/// <summary>
        /// Registers a SIP account for a SIP PBX.
        /// </summary>
        public virtual void Register()
        {
            phoneLine = softPhone.CreatePhoneLine(Account, new NatConfiguration(NatTraversalMethod.None));

            phoneLine.PhoneLineStateChanged += PhoneLine_PhoneLineInformationCanged;
            softPhone.IncomingCall += SoftPhone_IncomingCall;


            softPhone.RegisterPhoneLine(phoneLine);

        }

        /// <summary>
        /// Unregisters the registered phoneline from the SIP PBX. 
        /// </summary>
        public virtual void Unregister()
        {
            if (mediaGatewayAudioReceiver != null)
            {
                streamService.RemoveStream(mediaGatewayAudioReceiver.MediaStream);
                mediaGatewayAudioReceiver = null;
            }

            HangUp();

            softPhone.IncomingCall -= SoftPhone_IncomingCall;

            if (phoneLine != null)
            {
                phoneLine.PhoneLineStateChanged -= PhoneLine_PhoneLineInformationCanged;
                softPhone.UnregisterPhoneLine(phoneLine);
            }
        }

        #region This methods invoke client side methods.

        /// <summary>
        /// Indicates for the client, he can start and play his media streams. Invoke the client side OnInCall method.
        /// </summary>
        /// <param name="publishName"></param>
        /// <param name="playName"></param>
        void OnInCall(string publishName, string playName)
        {
            client.InvokeMethod("OnInCall", publishName, playName);
        }

        /// <summary>
        /// Indicates for the client, he has an incoming call. Invoke the client side OnIncomingCall method.
        /// </summary>
        /// <param name="phoneNumber">The request owner's phonenumber.</param>
        void OnIncomingCall(string phoneNumber)
        {
            client.InvokeMethod("OnIncomingCall", phoneNumber);
        }

        /// <summary>
        /// Indicates for the client, his phoneline state has changed.
        /// </summary>
        /// <param name="phoneLineInformation"></param>
        void OnPhoneLineStateChanged(PhoneLineState phoneLineInformation)
        {
            client.InvokeMethod("OnPhoneLineStateChanged", phoneLineInformation);
        }

        /// <summary>
        /// Indicates for the client, his call state has changed.
        /// </summary>
        /// <param name="callState"></param>
        void OnCallStateChanged(CallState callState)
        {
            client.InvokeMethod("OnCallStateChanged", callState);
        }
        /// <summary>
        /// Indicates for the client, his call request ended with error.
        /// </summary>
        /// <param name="callError"></param>
        void OnCallErrorOccurred(CallError callError)
        {
            client.InvokeMethod("OnCallErrorOccurred", callError);
        }

        /// <summary>
        /// Indicates for the client, gets a DTMF sign.
        /// </summary>
        /// <param name="dtmfsignal"></param>
        void OnCallDTMFReceived(DtmfSignal dtmfsignal)
        {
            client.InvokeMethod("OnDTMFReceived",dtmfsignal.Signal);
        }

        #endregion

       

        public void StreamPublished(IMediaStream mediaStream)
        {
            mediaGatewayAudioSender = new MediaGatewayAudioSender(mediaStream);
            mediaConnector.Connect(mediaGatewayAudioSender, phoneCallAudioSender);
        }

        public void StreamClosed()
        {
            if (mediaGatewayAudioSender != null)
            {
                mediaConnector.Disconnect(mediaGatewayAudioSender, phoneCallAudioSender);
                mediaGatewayAudioSender.Dispose();
                mediaGatewayAudioSender = null;
            }
        }
    }

}
